/*
 * JaamSim Discrete Event Simulation
 * Copyright (C) 2013 Ausenco Engineering Canada Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.jaamsim.probability;

import static org.junit.Assert.assertTrue;

import org.junit.Test;

import com.jaamsim.ProbabilityDistributions.ContinuousDistribution;
import com.jaamsim.Samples.SampleProvider;
import com.jaamsim.basicsim.ObjectType;
import com.jaamsim.events.EventManager;
import com.jaamsim.events.ProcessTarget;
import com.jaamsim.events.TestFrameworkHelpers;
import com.jaamsim.input.InputAgent;

public class TestContinuousDistribution {
	static class SampleDistribution extends ProcessTarget {
		final SampleProvider dist;
		final int numSamples;
		double total;

		public SampleDistribution(SampleProvider dist, int numSamples) {
			this.dist = dist;
			this.numSamples = numSamples;
		}
		@Override
		public String getDescription() {
			return "DistibutionUnitTest";
		}

		@Override
		public void process() {
			total = 0.0d;
			for (int i = 0; i < numSamples; i++)
				total += dist.getNextSample(0.0d);
		}
	}

	static double sampleDistribution(SampleProvider dist, int numSamples) {
		SampleDistribution target = new SampleDistribution(dist, numSamples);
		EventManager evt = new EventManager("DistibutionUnitTest");
		evt.clear();

		evt.scheduleProcessExternal(0, 0, false, target, null);
		TestFrameworkHelpers.runEventsToTick(evt, Long.MAX_VALUE, 100000000);
		return target.total;
	}

	@Test
	public void MeanAndStandardDeviation() {
		ObjectType t = InputAgent.defineEntityWithUniqueName(ObjectType.class, "TestType", "-", true);
		InputAgent.applyArgs(t, "JavaClass", "com.jaamsim.units.DimensionlessUnit");

		ContinuousDistribution dist = InputAgent.defineEntityWithUniqueName(ContinuousDistribution.class, "Dist", "-", true);
		InputAgent.applyArgs(dist, "UnitType", t.getName());
		InputAgent.applyArgs(dist, "ValueList", "1.0", "3.0", "5.0", "10.0");
		InputAgent.applyArgs(dist, "CumulativeProbabilityList", "0.0", "0.5", "0.8", "1.0");
		InputAgent.applyArgs(dist, "RandomSeed", "1");
		dist.validate();
		dist.earlyInit();

		int numSamples = 1000000;
		double total = TestContinuousDistribution.sampleDistribution(dist, numSamples);
		double mean = total / numSamples;

		assertTrue( Math.abs( dist.getSampleMean(0.0) - mean ) < 0.001 );
		assertTrue( Math.abs( dist.getSampleMean(0.0) / dist.getMeanValue(0.0) - 1.0 ) < 0.001 );
		assertTrue( Math.abs( dist.getSampleStandardDeviation(0.0) / dist.getStandardDeviation(0.0) - 1.0 ) < 0.001 );
	}
}
